mod array;
mod field;
mod function;
pub mod implementation;
mod instruction;
mod interface;
pub mod lib;
mod method;
pub mod object;
pub mod op_codes;
pub mod platform;
mod pointer;
pub mod r#type;
pub mod util;
pub mod value;
mod virtual_method;

use crate::vm::implementation::DvmImplementation;
use crate::vm::instruction::{Immediate, Instruction, Location, SourceCodeLocation};
use crate::vm::object::DvmObject;
use crate::vm::pointer::DvmPointer;
use crate::vm::value::DvmValue;
use function::DvmFunction;
use std::cell::RefCell;
use std::collections::HashMap;
use std::fmt::Display;
use std::rc::Rc;

fn dvm_panic(context: &DvmContext, message: &str) {
    println!("----");
    println!("Deimos Virtual Machine: Panic!");
    println!("{}", message);
    println!("----");
    println!("Current Instruction: {:?}", context.current_instruction);
    println!("Registers:");
    for (i, register) in context.registers.iter().enumerate() {
        println!("\tr{}: {:?}", i, register);
    }
    println!("Call stack:");
    for call in context.call_stack.iter() {
        println!("\t{}", call);
    }
    println!("Stack:");
    for (i, value) in context.stack.iter().enumerate() {
        println!("\t{}: {}", i, value);
    }
    println!("----");
}

macro_rules! dvm_panic {
    ($context: expr, $message: expr) => {
        dvm_panic($context, $message);
        panic!()
    };
}

pub type PlatformFunction = fn(context: &mut DvmContext);

pub struct DvmContext {
    call_stack: Vec<DvmCall>,
    current_instruction: Option<Instruction>,
    registers: Vec<DvmValue>,
    stack: Vec<DvmValue>,
    implementations: HashMap<String, Rc<DvmImplementation>>,
    static_functions: HashMap<String, Rc<DvmFunction>>,
    platform_functions: HashMap<String, PlatformFunction>,
}

impl DvmContext {
    pub fn new() -> Self {
        DvmContext {
            call_stack: vec![],
            current_instruction: None,
            registers: vec![],
            stack: vec![],
            implementations: HashMap::new(),
            static_functions: HashMap::new(),
            platform_functions: HashMap::new(),
        }
    }

    pub fn pop_stack(&mut self) -> DvmValue {
        self.stack.pop().unwrap_or_else(|| {
            dvm_panic!(self, "Stack underflow!");
        })
    }
}

pub struct DvmCall {
    function_name: Rc<String>,
    source_code_location: SourceCodeLocation,
}

impl Display for DvmCall {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "{} in {} @ {}:{}",
            self.function_name,
            self.source_code_location.source_file_name,
            self.source_code_location.line,
            self.source_code_location.char
        ))
    }
}

fn update_index(index: &mut usize, offset: isize) {
    if offset.is_negative() {
        *index -= offset.abs() as usize;
    } else {
        *index += offset as usize;
    }
}

fn push_call_frame(
    function_name: Rc<String>,
    source_code_location: SourceCodeLocation,
    context: &mut DvmContext,
) {
    context.call_stack.push(DvmCall {
        function_name,
        source_code_location,
    });
}

fn pop_call_frame(context: &mut DvmContext) {
    context.call_stack.pop();
}

fn compute_index_with_offset(
    registers: &Vec<DvmValue>,
    index_register: usize,
    offset: &Option<isize>,
) -> usize {
    let index = registers[index_register].as_usize().expect(&format!(
        "Could not convert the value of index_register {} to usize; found {:?}",
        index_register, registers[index_register]
    ));
    if let Some(o) = offset {
        if o.is_negative() {
            index - o.unsigned_abs()
        } else {
            index + o.unsigned_abs()
        }
    } else {
        index
    }
}

fn copy_from_source(context: &DvmContext, source: &Location) -> DvmValue {
    match source {
        Location::Register(register) => context.registers[*register].clone(),
        Location::Stack { offset } => context.stack[*offset].clone(),
        Location::Field {
            object_register,
            field_name,
        } => context.registers[*object_register].map_object(
            |o| o.borrow().read_field(field_name).clone(),
            |v| {
                dvm_panic!(context, &format!("Expected an object, but found: {}", v));
            },
        ),
        Location::Array {
            array_register,
            index_register,
            offset,
        } => {
            let index = compute_index_with_offset(&context.registers, *index_register, offset);
            context.registers[*array_register].map_array(
                |a| a.borrow().read_element(index),
                |v| {
                    dvm_panic!(context, &format!("Expected an array, but found: {}", v));
                },
            )
        }
    }
}

fn write_to_destination(context: &mut DvmContext, destination: &Location, value: DvmValue) {
    match destination {
        Location::Register(register) => {
            context.registers[*register] = value;
        }
        Location::Stack { offset } => {
            context.stack[*offset] = value;
        }
        Location::Field {
            object_register,
            field_name,
        } => {
            context.registers[*object_register]
                .expect_object()
                .borrow_mut()
                .write_field(field_name, value);
        }
        Location::Array {
            array_register,
            index_register,
            offset,
        } => {
            let index = compute_index_with_offset(&context.registers, *index_register, offset);
            context.registers[*array_register]
                .expect_array()
                .borrow_mut()
                .write_element(index, value);
        }
    }
}

fn immediate_to_value(immediate: &Immediate) -> DvmValue {
    match immediate {
        Immediate::Byte(b) => DvmValue::Byte(*b),
        Immediate::Int(i) => DvmValue::Int(*i),
        Immediate::Long(l) => DvmValue::Long(*l),
        Immediate::Double(d) => DvmValue::Double(*d),
        Immediate::Usize(s) => DvmValue::USize(*s),
        Immediate::Boolean(b) => DvmValue::Boolean(*b),
        Immediate::ConstantPointer {
            raw_pointer,
            length,
        } => DvmValue::ConstantPointer(DvmPointer {
            raw_pointer: *raw_pointer,
            length: *length,
        }),
        Immediate::Empty => DvmValue::Empty,
    }
}

// TODO: find all possible unwraps/expects and wrap with call to dvm_panic
pub fn run(instructions: &[Instruction], context: &mut DvmContext) {
    let mut index = 0;
    while index < instructions.len() {
        context.current_instruction = Some(instructions[index].clone());
        index += 1;
        match &instructions[index] {
            Instruction::MoveImmediate {
                destination,
                immediate,
            } => {
                write_to_destination(context, destination, immediate_to_value(immediate));
            }
            Instruction::Copy {
                source,
                destination,
            } => {
                let value = copy_from_source(context, source);
                write_to_destination(context, destination, value);
            }
            Instruction::Push { source_register } => {
                context
                    .stack
                    .push(context.registers[*source_register].clone());
            }
            Instruction::Pop {
                destination_register,
            } => {
                let value = context.stack.pop().expect("Stack underflow");
                context.registers[*destination_register] = value;
            }
            Instruction::AllocateObject {
                implementation_name,
                destination_register,
            } => {
                let implementation = context
                    .implementations
                    .get(implementation_name.as_str())
                    .unwrap();
                let object = DvmObject::new(implementation.clone());
                context.registers[*destination_register] =
                    DvmValue::Object(Rc::new(RefCell::new(object)));
            }
            Instruction::InvokeStatic {
                function_name,
                source_code_location,
            } => {
                let static_function = context
                    .static_functions
                    .get(function_name.as_str())
                    .unwrap()
                    .clone();
                push_call_frame(function_name.clone(), source_code_location.clone(), context);
                run(static_function.instructions(), context);
                pop_call_frame(context);
            }
            Instruction::InvokeObject {
                object_register,
                function_name,
                source_code_location,
            } => {
                let object = context.registers[*object_register].expect_object();
                let object_ref = object.borrow();
                let method = object_ref
                    .get_method(function_name.as_str())
                    .expect(&format!(
                        "Cannot find method {} for object {:?}",
                        function_name, object
                    ));
                let function = method.dm_fn();
                let instructions = function.instructions();
                push_call_frame(function_name.clone(), source_code_location.clone(), context);
                run(instructions, context);
                pop_call_frame(context);
            }
            Instruction::PlatformCall {
                platform_function_name,
                source_code_location,
            } => {
                let platform_function = context
                    .platform_functions
                    .get(platform_function_name.as_str())
                    .unwrap()
                    .clone();
                push_call_frame(
                    platform_function_name.clone(),
                    source_code_location.clone(),
                    context,
                );
                platform_function(context);
                pop_call_frame(context);
            }
            Instruction::Add { .. } => {}
            Instruction::Subtract { .. } => {}
            Instruction::Multiply { .. } => {}
            Instruction::Divide { .. } => {}
            Instruction::Modulo { .. } => {}
            Instruction::Power { .. } => {}
            Instruction::Jump { offset } => {
                update_index(&mut index, *offset);
            }
            Instruction::JumpEqual {
                left_register,
                right_register,
                offset,
            } => {
                let left_value = &context.registers[*left_register];
                let right_value = &context.registers[*right_register];
                if left_value == right_value {
                    update_index(&mut index, *offset);
                }
            }
            Instruction::JumpNotEqual { .. } => {}
            Instruction::JumpLessThan { .. } => {}
            Instruction::JumpGreaterThan { .. } => {}
            Instruction::JumpLessThanEqual { .. } => {}
            Instruction::JumpGreaterThanEqual { .. } => {}
            Instruction::Return => {
                break;
            }
        }
    }
}
